// INCLUDE: global uniform header

struct InstanceInput {
	@location(2) anchor: u32,
	@location(3) position: vec2<f32>,
	@location(4) angle: f32,
	@location(5) size: f32,
	@location(6) color_transform: vec4<f32>,
	@location(7) sprite_index: u32,
	@location(8) mask_index: vec2<u32>,
};

struct VertexInput {
	@location(0) position: vec3<f32>,
	@location(1) texture_coords: vec2<f32>,
}

struct VertexOutput {
	@builtin(position) position: vec4<f32>,
	@location(0) texture_coords: vec2<f32>,
	@location(1) texture_index: u32,
	@location(2) mask_coords: vec2<f32>,
	@location(3) mask_index: vec2<u32>,
	@location(4) color_transform: vec4<f32>,
}

@group(0) @binding(0)
var texture_array: binding_array<texture_2d<f32>>;
@group(0) @binding(1)
var sampler_array: binding_array<sampler>;


// INCLUDE: animate.wgsl
// INCLUDE: anchor.wgsl

@vertex
fn vertex_main(
	vertex: VertexInput,
	instance: InstanceInput,
) -> VertexOutput {


	let window_dim = global_data.window_size / global_data.window_scale.x;
	let scale = instance.size / window_dim.y;
	let sprite_aspect = global_sprites[instance.sprite_index].aspect;

	// Apply scale and sprite aspect
	// Note that our mesh starts centered at (0, 0). This is important!
	var pos: vec2<f32> = vec2(
		vertex.position.x * scale * sprite_aspect,
		vertex.position.y * scale
	);

	// Apply rotation (and adjust sprite angle, since sprites point north)
	pos = mat2x2(
		vec2(cos(instance.angle - 1.5708), sin(instance.angle - 1.5708)),
		vec2(-sin(instance.angle - 1.5708), cos(instance.angle - 1.5708))
	) * pos;

	// Correct for screen aspect, preserving height
	pos = vec2(
		pos.x / global_data.window_aspect.x,
		pos.y
	);

	pos = pos + anchor(
		instance.anchor,
		instance.position,
		vec2(instance.size * sprite_aspect, instance.size)
	);

	var out: VertexOutput;
	out.position = vec4<f32>(pos, 1.0, 1.0);
	out.color_transform = instance.color_transform;



	// TODO: function to get texture from sprite
	// Pick texture frame
	let t = global_atlas[u32(animate(instance.sprite_index, global_data.current_time.x, 0.0))];
	out.texture_index = u32(t.atlas_texture);
	out.texture_coords = vec2(t.xpos, t.ypos);
	if vertex.texture_coords.x == 1.0 {
		out.texture_coords = vec2(out.texture_coords.x + t.width, out.texture_coords.y);
	}
	if vertex.texture_coords.y == 1.0 {
		out.texture_coords = vec2(out.texture_coords.x, out.texture_coords.y + t.height);
	}

	// Pick mask image if mask is enabled
	// x coordinate of mask index is either 0 or 1, telling us whether or not to use a mask.
	// y coordinate is mask sprite index
	if instance.mask_index.x == 1u {
		let ms = global_sprites[instance.mask_index.y].first_frame;
		let m = global_atlas[ms];
		out.mask_index = vec2(1u, u32(m.atlas_texture));
		out.mask_coords = vec2(m.xpos, m.ypos);
		if vertex.texture_coords.x == 1.0 {
			out.mask_coords = vec2(out.mask_coords.x + m.width, out.mask_coords.y);
		}
		if vertex.texture_coords.y == 1.0 {
			out.mask_coords = vec2(out.mask_coords.x, out.mask_coords.y + m.height);
		}
	} else {
		out.mask_coords = vec2(0.0, 0.0);
		out.mask_index = vec2(0u, 0u);
	}


	return out;
}


@fragment
fn fragment_main(in: VertexOutput) -> @location(0) vec4<f32> {

	var mask: f32 = 1.0;
	if in.mask_index.x == 1u {
		mask = textureSampleLevel(
			texture_array[in.mask_index.y],
			sampler_array[0],
			in.mask_coords,
			0.0
		).a;
	}

	var color: vec4<f32> = textureSampleLevel(
		texture_array[in.texture_index],
		sampler_array[0],
		in.texture_coords,
		0.0
	).rgba * in.color_transform;

	// Apply mask and discard fully transparent pixels
	color = vec4(color.rgb, color.a *mask);
	if color.a == 0.0 {
		discard;
	}

	return color;
}